@@ -1,6 +1,8 @@
 """Storage Connectors for CloudWanderer."""
+from .base_connector import BaseStorageConnector
 from .dynamodb import DynamoDbConnector
 
 __all__ = [
-    'DynamoDbConnector'
+    'DynamoDbConnector',
+    'BaseStorageConnector'
 ]
